from tagc.data_pipe import (
    dataset_split,
    form_random_ds,
    sample_evaluation_from_path,
    xlsx_to_cases,
)
from tagc.data_utils import rawdata_stat
from tagc.io_utils import dump_json, load_datazip
from tagc.visualization import plot_tag_stat

CASE_NUM = 11418


def mk_cases(xlsx_p, case_p="cases.json"):
    cases = xlsx_to_cases(xlsx_p)
    assert len(cases) == CASE_NUM
    dump_json(case_p, cases)
    return case_p


def mk_unlabelled(
    final_dsp="standardDs.zip", cases_p="cases.json", outpath="unlabelled.json"
):
    sampled_cases = sample_evaluation_from_path(cases_p, final_dsp)
    dump_json(outpath, sampled_cases)
    return outpath


def mk_standardDs(final_dsp="standardDsTmp.zip", dst="out/standard", plot=True):
    dsps = dataset_split(final_dsp, dst)
    if plot:
        for idx, dsp in enumerate(dsps):
            rawdata = load_datazip(dsp)
            tag_stat = rawdata_stat(rawdata)
            tag_stat.to_csv(f"{dst}/data_stat{idx}.csv")
            fig = plot_tag_stat(tag_stat)
            fig.write_image(f"{dst}/data_stat{idx}.pdf")
    return dsps


def mk_randomDs(
    standard_dsps,
    eval_ret="data/evaluation/mona_j.csv",
    unlabelled_p="data/unlabelled.json",
    dst="out/random",
    plot=True,
):
    dsps = form_random_ds(standard_dsps, eval_ret, unlabelled_p, outdir=dst)
    if plot:
        for idx, dsp in enumerate(dsps):
            rawdata = load_datazip(dsp)
            tag_stat = rawdata_stat(rawdata)
            tag_stat.to_csv(f"{dst}/random_data_stat{idx}.csv")
            fig = plot_tag_stat(tag_stat)
            fig.write_image(f"{dst}/random_data_stat{idx}.pdf")
    return dsps



import json
import logging
import re
from functools import partial
from operator import methodcaller
from typing import List, NamedTuple, Optional, Tuple

from toolz.functoolz import pipe

logging.basicConfig()
logger = logging.getLogger("cleaner")

_FIELDS = [
    "PARTICLES:",
    "MEGAKARYOCYTES",
    "EXTRINSIC",
    "ERYTHROPOIESIS",
    "GRANULOPOIESIS",
    "RETICULUM",
    "HEMOSIDERIN",
    "CELLULARITY",
    "COMMENT:",
    "INTERPRETATION",
    "CONCLUSIONS",
]

MAX_KEY_LEN = 59
MIN_KEY_LEN = 5
MIN_VALUE_LEN = 3
ENTRY_LEN = 2


class Entry(NamedTuple):
    key: str
    value: str


class Cleaner(object):
    def __init__(self):
        self._legit = 0
        self._warning = 0
        self._warning_collector = set()

    @property
    def not_used_texts(self):
        return self._warning_collector

    def parse_aspirate_result(self, aspirate_result_text: str) -> dict:
        return pipe(aspirate_result_text, self._clean_text, self._parse_text)

    def _clean_text(self, aspirate_result_text: str) -> str:
        return pipe(
            aspirate_result_text,
            _remove_signature,
            _remove_inline_space,
            _remove_end_space,
            _remove_reporting_system,
            _replace_add,
        )

    def _parse_text(self, aspirate_result_cleaned_text: str) -> dict:
        return pipe(aspirate_result_cleaned_text, self._entry_divide, self._form_dict)

    def _entry_divide(self, text: str) -> List[str]:
        for term in _FIELDS:
            text = text.replace(term, "\n" + term, 1)
        text = _special_field(text)
        return text.split("\n")

    def _form_dict(self, entry_texts: List[str]) -> dict:
        return {
            entry.key: entry.value
            for entry in map(self._extract_entry, entry_texts)
            if entry is not None
        }

    def _extract_entry(self, entry_text: str) -> Optional[Entry]:
        sudo_entry = tuple(map(str.strip, entry_text.split(":", 1)))
        entry = None
        if Cleaner._is_legit_entry(sudo_entry):
            entry = self._form_entry(sudo_entry)
        if Cleaner._deserve_warning(entry, sudo_entry):
            self._process_warning_case(entry_text)
        return entry

    @staticmethod
    def _is_legit_entry(sudo_entry: Tuple[str, ...]) -> bool:
        return (
            len(sudo_entry) == ENTRY_LEN
            and _nonword_in_attr(sudo_entry[0]) is None
            and sudo_entry[1] != ""
            and MIN_KEY_LEN < len(sudo_entry[0]) < MAX_KEY_LEN
            and MIN_VALUE_LEN < len(sudo_entry[1])
        )

    def _form_entry(self, sudo_entry: tuple):
        self._legit += 1
        return Entry(sudo_entry[0], sudo_entry[1])

    @staticmethod
    def _deserve_warning(entry, sudo_entry: tuple):
        return entry is None and len(sudo_entry) != 1 and sudo_entry[1] != ""

    def _process_warning_case(self, entry_text: str):
        self._warning += 1
        self._warning_collector.add(entry_text)

    def parse_as_json(self, aspirate_result_text: str, out_path: str):
        with open(out_path, "w") as json_:
            json.dump(self.parse_aspirate_result(aspirate_result_text), json_)
        self.job_report()

    def job_report(self):
        warn_rate = self._warning / (self._legit + self._warning)
        print(
            f"Parser: {self._legit} legit fields,{self._warning}"
            f"({warn_rate : .3f}) warning"
        )


_remove_inline_space = partial(re.compile(r"\s+").sub, " ")
_remove_end_space = partial(re.compile(r"\s$").sub, "")
_nonword_in_attr = re.compile(r"\d+|\.|@+|TO|OF").search
_remove_signature = partial(re.compile(r"REPORTING PHYSICIAN.*|D[rR][.]? .*").sub, "")
_replace_add = methodcaller("replace", "&", "AND")
_remove_reporting_system = partial(re.compile(r"\(REPORTING.*\)").sub, "")
_special_field = partial(re.compile(r"PERIPHERAL (?!BLOOD[\s.])").sub, "\nPERIPHERAL ")


def has_content(case: dict):
    return len(case) != 0



import os

import pandas as pd
from sklearn.preprocessing import MultiLabelBinarizer

from tagc.data_utils import rawdata_stat
from tagc.io_utils import (
    build_eval_json,
    dump_json,
    dump_state,
    load_datazip,
    load_json,
)
from tagc.mask_explain import MaskExplainer, filter_top, top_keywords
from tagc.model import StandaloneModel, label_output
from tagc.validation import (
    dimension_reduction,
    eval_model,
    get_tag_states,
    get_unlabelled_state,
)
from tagc.visualization import kw_plot, plot_tag_stat, state_plot


def make_figures(model_p: str, dsp: str, unlabelled_p: str, dst="figF"):
    dataset = load_datazip(dsp)
    model = StandaloneModel.from_path(model_p, keep_key=True, max_len=150)
    over = 5
    mlb = MultiLabelBinarizer().fit(dataset.y_tags)
    os.makedirs(dst, exist_ok=True)
    # Rawdata_stat
    fn = f"{dst}/data_stat.csv"
    if os.path.exists(fn):
        tag_stat = pd.read_csv(fn, index_col=0)
    else:
        tag_stat = rawdata_stat(dataset)
        tag_stat.to_csv(fn)
    fig = plot_tag_stat(tag_stat)
    fig.write_image(f"{dst}/data_stat.pdf")

    # Unlabelled
    fn = f"{dst}/unlabel_tsne.csv"
    if os.path.exists(fn):
        unstate_df = pd.read_csv(fn, index_col=0)
    else:
        sampled_cases = load_json(unlabelled_p)
        sampled_state = get_unlabelled_state(model, sampled_cases, mlb)
        dump_state(sampled_state, state_p=f"{dst}/unstate.pkl")
        unstate_df = dimension_reduction(sampled_state, "TSNE", n_components=2)
        unstate_df.to_csv(fn)
        preds = model.over_predict(sampled_cases, n=over)
        thresh_items = label_output(preds)
        pred_prob = [list(zip(mlb.classes_, pred)) for pred in preds]
        eval_json = build_eval_json(sampled_cases, pred_prob, thresh_items)
        dump_json(f"{dst}/eval.json", eval_json)
    fig = state_plot(unstate_df, 12)
    fig.write_image(f"{dst}/unlabel_tsne.pdf")
    fig.write_html(f"{dst}/unlabel_tsne.html")

    # Labelled
    fn = f"{dst}/label_tsne.csv"
    if os.path.exists(fn):
        state_df = pd.read_csv(fn, index_col=0)
    else:
        states = get_tag_states(model, dataset, mlb)
        state_df = dimension_reduction(states, "TSNE", n_components=2)
        state_df.to_csv(fn)
    fig = state_plot(state_df, 12)
    fig.write_image(f"{dst}/label_tsne.pdf")
    fig.write_html(f"{dst}/label_tsne.html")

    # Performance
    eval_model(model, dataset, over, mlb, dst, "")

    # Top key
    fn = f"{dst}/top_key.json"
    if os.path.exists(fn):
        top_key = load_json(fn)
    else:
        maskExplainer = MaskExplainer(mlb)
        top = top_keywords(maskExplainer, model, dataset.x_dict)
        top_key = filter_top(top, dataset.y_tags, thresh=20)
        dump_json(fn, top_key)
    fig = kw_plot(top_key)
    fig.write_image(f"{dst}/knockout_result.pdf")


if __name__ == "__main__":
    from fire import Fire

    Fire(make_figures)
